import logging
import tempfile
import unittest, os
# ======================================================================================================================
from multiprocessing import Process

from fragmenstein.igor.pyrosetta_import import pyrosetta

# ======================================================================================================================
from rdkit import Chem
from rdkit.Chem import AllChem, rdqueries
from rdkit.Geometry import Point3D

# TESTS IS EXTERNAL TO FRAGMENSTEIN DO NOT CHANGE TO RELATIVE!
from fragmenstein import Monster, Igor, Victor
from fragmenstein.demo import Mac1
from fragmenstein.error import FragmensteinError
from fragmenstein.mpro import MProVictor
from typing import *
import numpy as np
import json
from rdkit import Chem, rdBase
from fragmenstein import Laboratory, Victor

# ======================================================================================================================

class Internals(unittest.TestCase):

    def setUp(self):
        Victor.capture_rdkit_log()
    def test_triangle(self):
        """
        Test triangle prevention.
        """
        probanda = Chem.MolFromSmiles('CCC')
        AllChem.EmbedMolecule(probanda)
        monster = Monster([probanda])
        zeroth = probanda.GetAtomWithIdx(0)
        second = probanda.GetAtomWithIdx(2)
        # 0 - 2 would make a triangle
        self.assertTrue(
            monster._is_would_be_triangle(zeroth, second))  # connecting zeroth and second would make a triangle
        self.assertEqual(monster._get_triangle(zeroth, second), 1)  # connecting 0+2, would make 1 the vertex.

    def make_mol(self, smiles: str) -> Chem.Mol:
        mol = Chem.MolFromSmiles(smiles)
        dummies = mol.GetAtomsMatchingQuery(rdqueries.AtomNumEqualsQueryAtom(0))
        for dummy in dummies:
            dummy.SetAtomicNum(6)  # carbon is 6 not 12!
        AllChem.EmbedMolecule(mol)
        for dummy in dummies:
            dummy.SetAtomicNum(0)
        return mol

    def make_pair_by_split(self, conjoined: Chem.Mol, atom_idx: int) -> Tuple[Chem.Mol, Chem.Mol]:
        # make overlapping mols by getting a single molecule, and split it
        # this gives more control over Chem.rdMolAlign.AlignMol as this may overlap other atoms.
        # negative weights does not work...
        # fore
        bond = conjoined.GetBondBetweenAtoms(atom_idx, atom_idx + 1)
        fragged = Chem.FragmentOnBonds(conjoined, [bond.GetIdx()], addDummies=False)
        fore = Chem.GetMolFrags(fragged, asMols=True)[0]
        bond = conjoined.GetBondBetweenAtoms(atom_idx - 1, atom_idx)
        fragged = Chem.FragmentOnBonds(conjoined, [bond.GetIdx()], addDummies=False)
        aft = Chem.GetMolFrags(fragged, asMols=True)[1]
        return fore, aft

    def test_merge_on_same_dummy(self):
        conjoined = self.make_mol('O=C(O)CSCC#N')
        acetyl, nitrile = self.make_pair_by_split(conjoined, 4)
        # merge
        monster = Monster([acetyl, nitrile])
        merger = monster.simply_merge_hits()
        dummies = merger.GetAtomsMatchingQuery(Chem.rdqueries.AtomNumEqualsQueryAtom(0))
        for dummy in dummies:
            self.assertEqual(len(dummy.GetNeighbors()), 1)
        self.assertEqual(len(Chem.GetMolFrags(merger)), 1)

    def translate(self, mol, x=0, y=0, z=0):
        """
        Translates the molecule in place

        :param mol:
        :param x:
        :param y:
        :param z:
        :return:
        """
        translation = np.array([[1, 0, 0, x],
                                [0, 1, 0, y],
                                [0, 0, 1, z],
                                [0, 0, 0, 1]], dtype=np.double)
        AllChem.TransformConformer(mol.GetConformer(0), translation)

    def test_join(self):
        wanted = 'c1ccc(Oc2ccccc2)cc1'
        benzene = Chem.MolFromSmiles('c1ccccc1')
        AllChem.EmbedMolecule(benzene)
        moved = Chem.Mol(benzene)
        self.translate(moved, x=5)
        mol = Monster([benzene, moved]).combine().positioned_mol
        found = Chem.MolToSmiles(Chem.RemoveHs(mol))
        self.assertEqual(wanted, found, 'The joining differs')

    def test_distance(self):
        methane = Chem.MolFromSmiles('C')
        AllChem.EmbedMolecule(methane)
        ammonia = Chem.MolFromSmiles('N')
        AllChem.EmbedMolecule(ammonia)
        self.translate(ammonia, x=3)
        monster = Monster([methane, ammonia])
        monster.combine(keep_all=False, joining_cutoff=2)
        self.assertEqual(1, len(monster.unmatched), 'discard error')
        monster.combine(keep_all=True, joining_cutoff=5)
        self.assertEqual(0, len(monster.unmatched), 'discard error')
        try:
            monster.combine(keep_all=True, joining_cutoff=2)
            self.fail('should have raised a connection error')
        except FragmensteinError as error:
            pass

    def test_neigh_bonding(self):
        """
        This was to address the bug from 5/1/23.

        Issue was in ``_nan_fill_submatrix``, which was not blanking due to
        ``isinstance(np.int64(42), int) == False``
        """
        # make mols
        methane: Chem.Mol = Chem.MolFromSmiles('C')
        methane.SetProp('_Name', 'methane')
        AllChem.EmbedMolecule(methane, coordMap={0: Point3D(0, 0, 0)})
        ammonia = Chem.MolFromSmiles('N')
        ammonia.SetProp('_Name', 'ammonia')
        # merge
        monster = Monster([])
        AllChem.EmbedMolecule(ammonia, coordMap={0: Point3D(3, 0, 0)})
        self.translate(ammonia, x=3)
        # monster.join_neighboring_mols(methane, ammonia)
        combo, candidates = monster._find_all_closest(methane, ammonia)  # _find_all_closest is in communal
        self.assertEqual(combo.GetNumAtoms(), 2)
        self.assertNotEqual(candidates[0][0], candidates[0][1])

    def test_percent_hybrid(self):
        chlorobutane = Chem.MolFromSmiles('[Cl]CCCC')
        origins = [[], ['ethanol.1'], ['ethanol.2', 'isopronanol.3'], ['isopronanol.2'], ['isopronanol.4']]
        chlorobutane.SetProp('_Origins', json.dumps(origins))
        hybrid = Laboratory.percent_hybrid(None, chlorobutane)
        self.assertEqual(hybrid, 34)  # ethanol has 1 single origin atom, isopronanol has 2

    def test_assign_empty(self):
        vicky = Victor.__new__(Victor)
        vicky.apo_pdbblock = Mac1.get_template()
        print(vicky._get_empty_resi())



    # def test_doubleconstraint(self):
    #     diaminopentane = Chem.MolFromSmiles('NCCCCCN')
    #     AllChem.EmbedMolecule(diaminopentane)
    #     AllChem
    #
    #     #Monster([])

if __name__ == '__main__':
    unittest.main()



import unittest
from .internals import Internals
from .problems import UnresolvedProblems
from .mapping import Mappings
from .monster_place import MonsterPlaceTests
from .monster_combine import MonsterCombineTests
from .MPro_combine import VictorCombineTests
from .MPro_place import MProPlaceTester
from .multivictor import MultivictorPlaceTests
from .walton import WaltonTests



from rdkit import Chem
from rdkit.Chem import BRICS, AllChem
from typing import List, Tuple, Any, Dict
import numpy as np
import pandas as pd

class AccountableBRICS:
    """
    BRICS decomposition that remembers where stuff came from.

    This stores the original inspiration as an isotope label
    It is potentially dangerous as dummy atom isotope number plays a role too.
    No property is kept however, so it is the only working solution.
    The attachment atom is labelled in the built molecule with `bridging_atom` which contains its inspiration's name

    ... code-block:: python
        decomposer = AccountableBRICS(hits)
        df: pd.DataFrame = decomposer(cutoff=decomposer.median, max_mergers=100)
        decomposer.info
        # {'N_hits': 44, 'N_fragments': 157, 'too_small': 0, 'missing': [], 'max_reached': True, 'N_built': 100}

    The output of ``__call__`` is a pandas dataframe with the following columns:

    - ``name``: the name of the built molecule ('build#👾')
    - ``smiles``: the smiles of the built molecule
    - ``hits``: the hits used to build the molecule
    - ``built_molecule``: the built molecule

    These can be passed to ``Laboratory.place`` to place the built molecules.

    ... code-block:: python
        pyrosetta.init(extra_options=extra_options)
        lab = Laboratory(pdbblock=👾👾👾, covalent_resi=None)
        placements:pd.DataFrame = lab.place(df, expand_isomers=False, n_cores=12)
        display(placements)
    """

    def __init__(self, hits: List[Chem.Mol]):
        self.hits = {}
        for i, hit in enumerate(hits):
            hit.SetIntProp('decompose_index', i + 1000)
            self.hits[i + 1000] = hit
        self.info = {'N_hits': len(self.hits),
                     'N_fragments': 0,
                     'too_small': 0,
                     'missing': [],
                     'max_reached': False}
        self.fragments: List[Chem.Mol] = []

    def decompose(self, mol: Chem.Mol) -> List[Chem.Mol]:
        name = mol.GetProp('_Name')
        index = mol.GetIntProp('decompose_index')
        for atom in mol.GetAtoms():
            atom.SetProp('ori_name', name)
            atom.SetIntProp('ori_i', atom.GetIdx())
        fragments = list(BRICS.BRICSDecompose(mol, keepNonLeafNodes=False, returnMols=True))
        # label
        for i, frag in enumerate(fragments):
            # this is a waste of time:
            frag.SetProp('_Name', f'{name}#{i}')
            for atom in frag.GetAtoms():
                atom.SetProp('ori_name', name)
            # but isotope is kept:
            dummy: Chem.Atom
            for dummy in frag.GetAtomsMatchingQuery(AllChem.AtomNumEqualsQueryAtom(0)):
                dummy.GetNeighbors()[0].SetIsotope(index)
                pass
        return fragments

    def __call__(self, cutoff: int = 0, max_mergers: int = 10_000, maxDepth: int = 4) -> pd.DataFrame:
        self.fragments = []
        for hit in self.hits.values():
            self.fragments.extend(self.decompose(hit))
        self.info['N_fragments'] = len(self.fragments)
        mergers: List[Chem.Mol] = []
        accepted: int = 0
        results: List[Dict[str, Any]] = []
        for built in BRICS.BRICSBuild(self.fragments,
                                      onlyCompleteMols=True,
                                      maxDepth=maxDepth):
            if built.GetNumHeavyAtoms() < cutoff:
                self.info['too_small'] += 1
                continue
            accepted += 1
            inspirations: List[Chem.Mol] = self.get_inspirations(built)
            results.append({'name': f'build#{len(results)}',  # required by Laboratory
                            'built_molecule': built,
                            'smiles': Chem.MolToSmiles(built),  # required by Laboratory
                            'hits': inspirations,
                            })
            if accepted >= max_mergers:
                self.info['max_reached'] = True
                break
        self.info['N_built'] = len(results)
        return pd.DataFrame(results)

    @property
    def median(self):
        return np.median(list(map(Chem.Mol.GetNumHeavyAtoms, self.hits.values())))

    def get_inspirations(self, built: Chem.Mol) -> List[Chem.Mol]:
        inspirations: List[Chem.Mol] = []
        atom: Chem.Atom
        for atom in built.GetAtomsMatchingQuery(AllChem.IsotopeGreaterQueryAtom(0)):
            i = atom.GetIsotope()
            if i not in self.hits:
                self.info['missing'].append(i)
                continue
            inspirations.append(self.hits[i])
            atom.SetIsotope(0)
            atom.SetProp('bridging_atom', self.hits[i].GetProp('_Name'))
        return sorted(inspirations, key=Chem.Mol.GetNumHeavyAtoms, reverse=True)


import logging
import pebble
import operator
from typing import (Any, Callable, Union, Iterator, Sequence, List, Dict)
from collections import Counter
import pandas as pd
from rdkit import Chem
from ..monster import Monster
from ..victor import Victor

# not needed for binarize... but just in case user is not using them...
Chem.SetDefaultPickleProperties(Chem.PropertyPickleOptions.AllProps)

def binarize(mol:Chem.Mol, ignore_errors:bool=True) -> bytes:
    # this is convoluted as None is a common error outcome with RDKit
    # so making it less cryptic
    if isinstance(mol, bytes):
        return mol   # messed up. it was bytes.
    elif not isinstance(mol, Chem.Mol): # likely None
        if ignore_errors:
            mol = Chem.Mol()    # noqa
        elif mol is None:
            raise TypeError('A Chem.Mol was expected by got None. '+
                            'RDKit returns None if it fails an operation.')
        else:
            raise TypeError('the molecule provide is not a Chem.Mol, '+
                            f'but a {type(mol).__name__}')
    exception = Exception if ignore_errors else ()
    try:
        return mol.ToBinary(propertyFlags=0b00010111)    # noqa it's an instance.
    except KeyboardInterrupt as err:
        raise err
    except KeyboardInterrupt as err:
        raise err
    except exception as error:
        return Chem.Mol().ToBinary(propertyFlags=0b00010111)  # noqa


def unbinarize(bin: bytes, ignore_errors:bool=True) -> Union[Chem.Mol, None]:
    exception = Exception if ignore_errors else ()
    if isinstance(bin, Chem.Mol):  # lol
        return bin
    elif not isinstance(bin, bytes) and ignore_errors:
        return None
    elif not isinstance(bin, bytes):
        raise TypeError('the molecule binary provide is not a bytes')
    try:
        return Chem.Mol(bin)
    except KeyboardInterrupt as err:
        raise err
    except exception:
        return None

class LabBench:

    # the ``outcome`` column in the pandas dataframe can have these values in order of niceness:
    category_labels = ['crashed', 'too distant', 'timeout', 'unstable', 'equally sized', 'deviant', 'acceptable']

    def __init__(self, pdbblock: str,
                 covalent_resi: Union[int, str, None] = None,
                 ligand_resi: Union[str, None]=None,
                 run_plip: bool=False,
                 ):
        self.pdbblock = pdbblock
        self.covalent_resi = covalent_resi
        self.init_options = '-ex1 -ex2 -no_optH false -mute all -ignore_unrecognized_res true -load_PDB_components false'
        self.raw_results = []
        self.ligand_resi = ligand_resi
        self.Victor = Victor  # So it can be swapped for a subclass w/o the need to subclass Laboratory
        self.run_plip = run_plip
        self.blacklist = []  # list of names to skip
        if not len(Victor.journal.handlers):
            Victor.enable_stdout(logging.CRITICAL)

    # keep it just in case: (not called within class as that would require `self.__class__.binarize`)
    binarize: Callable[[Chem.Mol, bool], bytes] = staticmethod(binarize)
    unbinarize: Callable[[bytes, bool], Union[Chem.Mol, None]] = staticmethod(unbinarize)

    def get_completed(self, futures: pebble.ProcessMapFuture) -> pd.DataFrame:
        """
        A pebble.ProcessMapFuture cannot be stored as an instance attribute as it's not picklable
        as it handles the lock. As a result it must be passed.

        Fills ``self.raw_results`` with the results of the futures, before returning a dataframe.
        """
        result_iter: Iterator[int] = futures.result()
        self.raw_results = []
        while True:
            try:
                result = next(result_iter)
                self.raw_results.append(result)
            except TimeoutError as error:
                Victor.journal.error("Function took longer than %d seconds" % error.args[1])
                self.raw_results.append({'error': 'TimeoutError', 'name': ''})
            except StopIteration as error:
                # it would be nice having a mock entry with the expected values...
                # self.raw_results.append({})
                break
            except KeyboardInterrupt as error:
                print('Keyboard!')
                self.raw_results.append({'error': 'KeyboardInterrupt', 'name': ''})
                break
            except Exception as error:
                Victor.journal.error(f'{error.__class__.__name__}: {error}')
                self.raw_results.append({'error': error.__class__.__name__, 'name': ''})
        # list of dict to dataframe
        df = pd.DataFrame(self.raw_results)
        if not len(df) or '∆∆G' not in df.columns:
            Victor.journal.critical('No results were found. Returning an empty dataframe.')
            return df
        df['LE'] = df.apply(
            lambda row: - row['∆∆G'] / (row.N_constrained_atoms + row.N_unconstrained_atoms),
            axis=1)
        nan_to_list = lambda value: value if isinstance(value, list) else []
        df['disregarded'] = df.disregarded.apply(nan_to_list)  # str
        df['regarded'] = df.regarded.apply(nan_to_list)  # str
        df['unminimized_mol'] = df.unmin_binary.apply(unbinarize)
        df['minimized_mol'] = df.min_binary.apply(unbinarize)
        df['hit_mols'] = df.hit_binaries.apply(lambda l: [unbinarize(b) for b in l] if isinstance(l, Sequence) else [])
        df['percent_hybrid'] = df.unminimized_mol.apply(self.percent_hybrid)
        return df

    def categorize(self,
                   row: pd.Series,
                   size_tolerance: int=0,
                   move_cutoff:float=1.,
                   ddG_cutoff:float=0.,
                   ) -> str:
        """
        Given a row categorise the 'outcome' field.
        Called by ``get_completed``.

        size_tolerance is the number of atoms below the largest hit that is still acceptable.
        move_cutoff is the RMSD of the minimized molecules. Below this is acceptable.
        ddG_cutoff is the ∆∆G of the minimized molecules. Below this is acceptable.

        Subclass Laboratory to change what the outcome field ends up being.
        But do remember to update the ``category_labels`` attribute.
        """
        # see category_labels for list of values.
        is_filled: Callable[[Any], int] = lambda value: len(value) != 0 if hasattr(value, '__len__') else False
        if 'disregarded' not in row.index:
            return 'crashed'  # the whole row is empty
        elif is_filled(row.disregarded) or 'DistanceError' in row.error:
            # either is the same thing, but the Victor error catching makes them different
            return 'too distant'
        elif 'TimeoutError' in row.error:
            return 'timeout'
        elif is_filled(row.error):
            return 'crashed'
        elif max(map(operator.methodcaller('GetNumHeavyAtoms'), row.hit_mols)) - size_tolerance  \
                >= row.unminimized_mol.GetNumHeavyAtoms():
            return 'equally sized'
        elif row.comRMSD > move_cutoff:
            return 'too moved'
        elif row['∆∆G'] >= ddG_cutoff:
            return 'too contorted'
        else:
            return 'acceptable'

    def percent_hybrid(self, mol: Chem.Mol) -> float:
        """
        Given the origins how much of the molecule is solely from the second hit?
        It's the ratio of the number of atoms from the second hit to the total number of atoms with an inspiration.
        Do note that monster has the dynamic attributes, ``.percent_common`` and ``.num_common``.
        This is how-much and how-many of the molecule is common to both hits.
        This metric is not in Monster but in the Laboratory, making it handy for overriding.
        It also aims at not diagnosing how much overlap there is but to diagnose how combined the molecule is.

        Basically there are a bunch of Qs to ask if there are two hits:

        * How much of a molecule is common to both?
        * How much is from either single hit?
        * How much is unmapped?

        But when discussing mergers,
        if a bunch of atoms are common, the first contribute more,
        but the second hit does not contribute anything
        then it's not a merger: it's just the first hit.
        Hence the need for this odd ratio.
        """
        if not isinstance(mol, Chem.Mol):
            return float('nan')
        # origins: list of one entry for each followup atom to list of strings of hit name dot atom name
        origins: List[List[str]] = Monster.origin_from_mol(None, mol)
        hit_names = [[ori_name.split('.')[0] for ori_name in atomic if ori_name] if atomic else [] for atomic in origins]
        flat_names = [name for atomic in hit_names for name in atomic]
        names = set(flat_names)
        if len(names) == 0:  # no inspirations
            return float('nan')
        elif len(names) == 1:  # one inspirations
            return 0.
        # hit to number of atoms with only that hit as origin
        single_origin: Dict[str, int] = {name: sum([name in atomic for atomic in hit_names if len(atomic) == len(names) - 1]) for name in names}
        sorted_names = sorted(single_origin, key=single_origin.get, reverse=True)
        if sum(single_origin.values()) == 0:
            return 0.
        return 100 - int(single_origin[sorted_names[0]] / sum(single_origin.values()) * 100)

    def __call__(self,
                 iterator: Iterator,
                 fun: Callable,
                 n_cores: int = 4,
                 timeout: int = 240,
                 max_tasks: int = 0,  # 0 mean infinity
                 asynchronous: bool = False
                 ):
        """Combine/permute the molecules ``mols``
        on ``n_cores`` subprocesses.
        killing any that live longer than ``timeout`` seconds.
        The method returns an iterator of promises ``pebble.ProcessMapFuture`` if ``asynchronous`` is True,
        or the results as a pandas DataFrame. To convert the promises to a dataframe use ``get_completed``."""

        def max_out(inner_iterator, maximum: int):
            for i, item in zip(range(maximum), inner_iterator):
                yield item

        if max_tasks > 0:
            iterator = max_out(iterator, max_tasks)

        with pebble.ProcessPool(max_workers=n_cores, max_tasks=n_cores) as pool:
            futures: pebble.ProcessMapFuture = pool.map(fun, iterator, timeout=timeout)
        if asynchronous:
            return futures
        else:
            return self.get_completed(futures)



from typing import Sequence
from rdkit import Chem
from rdkit.Chem import AllChem, Draw
from IPython.display import display
from matplotlib.colors import ColorConverter
import nglview as nv
from nglview.component import ComponentViewer
from io import StringIO


def display_mols(mols: Sequence[Chem.Mol],
                 molsPerRow=5,
                 subImgSize=(150, 150),
                 useSVG=True):
    """
    Generic wrapper for calling ``display(Draw.MolsToGridImage)``
    """
    if not mols:
        return  # no mols, no display
    flattos = [AllChem.RemoveHs(mol) for mol in mols if isinstance(mol, Chem.Mol)]
    for mol in flattos:
        AllChem.Compute2DCoords(mol)
    dopts = Draw.rdMolDraw2D.MolDrawOptions()  # noqa
    dopts.addAtomIndices = True
    hex_map = {atom.GetIdx(): atom.GetProp('_color') for atom in mol.GetAtoms()
               if atom.HasProp('_color') and atom.GetProp('_color')}
    rgb_map = {i: ColorConverter().to_rgb(n) for i, n in hex_map.items()}
    dopts.highlightAtomColors= list(rgb_map.values())
    dopts.highlightAtoms= list(rgb_map.keys())
    display(Draw.MolsToGridImage(flattos,
                                 legends=[mol.GetProp('_Name') if mol.HasProp('_Name') else '-' for mol in mols],
                                 subImgSize=subImgSize,
                                 useSVG=useSVG,
                                 molsPerRow=molsPerRow,
                                 drawOptions=dopts))

class MolNGLWidget(nv.NGLWidget):
    """
    Adds:

    a method ``add_mol`` that simply adds an rdkit molecule to the viewer.

    .. code-block:: python

        from rdkit import Chem
        mol = Chem.MolFromSmiles('CC')
        view = MolNGLWidget()
        view.add_mol(mol, colorValue='pink')

    A method ``add_neighbors`` that adds a representation of the neighbours of a given selection.

    .. code-block:: python

        view.add_neighbors('[ATP]', comp_id=0, radius=5, style='hyperball', color='gainsboro')
    And a methdo
    """

    def add_mol(self, mol: Chem.Mol, colorValue: str = '') -> ComponentViewer:
        """
        Add a rdkit.Chem to an NGLWidget.
        This function is used by Walton and Monster

        :param mol: rdkit.Chem.Mol
        :param colorValue: if blank, the color in the property _color is used.
        :return:
        """
        if not mol:
            raise ValueError( 'Provided mol is None: '+\
                              'if user manual tinkering happened, please run monster.fix_hits')
        fh = StringIO(Chem.MolToPDBBlock(mol))  # I want atom names if present
        comp: ComponentViewer = self.add_component(fh,  # noqa it's there.
                                                                name=mol.GetProp('_Name'),
                                                                ext='pdb'
                                                                )
        comp.remove_ball_and_stick()
        if not colorValue and mol.HasProp('_color'):
            colorValue = mol.GetProp('_color')
        if colorValue:
            comp.add_representation('ball+stick', colorValue=colorValue, multipleBond=True)
        else:
            comp.add_representation('ball+stick', multipleBond=True)
        # _color business stems from Walton.
        return comp

    def remove_all_components(self):
        """
        This removes the components and all traces of their existance in the widget

        :return:
        """
        self._js(f"""this.stage.removeAllComponents()""")
        i = 0
        while hasattr(self, f'component_{i}'):
            delattr(self, f'component_{i}')
            i += 1
        self._ngl_component_ids = []

    def add_neighbors(self, selection: str, comp_id: int = 0, radius: float = 5, style: str = 'hyperball',
                      color: str = 'gainsboro'):
        """
        Given a compounent id as an integer interpreted by NGL JS (not 'component_0'), show the neighbours.

        :param selection: NGL style selection string, e.g. ``[ATP] or 1-30:A``
        :param comp_id: integer not string
        :param radius: float of atom to atom distance max
        :param style: see NGL manual for the representation styles. Default is ``hyperball``
        :param color: hashtag prefixed hex color string or CSS color name
        :return:
        """
        self._js(f"""const comp = this.stage.compList[{comp_id}]
                     const target_sele = new NGL.Selection('{selection}');
                     const radius = {radius};
                     const neigh_atoms = comp.structure.getAtomSetWithinSelection( target_sele, radius );
                     const resi_atoms = comp.structure.getAtomSetWithinGroup( neigh_atoms );
                     comp.addRepresentation( "{style}", {{sele: resi_atoms.toSeleString(),
                     									  colorValue: "{color}",
                                                          multipleBond: true
                                                                       }});
                     comp.addRepresentation( "contact", {{sele: resi_atoms.toSeleString()}});
                """)

    def add_selection_signal(self, molname_id: str, atom_id: str):
        """
        Add a signal to the viewer than fills the elements #molname_id and #atom_id
        when user clicks on an atom —not bond.
        
            
        .. code-block:: python
            
            from IPython.display import display, HTML
            # in real code the html entities &gt;/&lt; would be actual greater-than and lesser-than signs
            display(HTML(f'&gt;div id="#{molname_id}"&lt;&gt;/div>&gt;div id="#{atom_id}"&lt;&gt;/div&lt;'))
            view.add_selection_signal(molname_id, atom_id)

        :param molname_id:
        :param atom_id:
        :return:
        """
        self._js(f'''this.stage.signals.clicked.add(pickingProxy => {{
                if (pickingProxy && (pickingProxy.atom || pickingProxy.bond )){{
                    const atom = pickingProxy.atom || pickingProxy.closestBondAtom;
                    const component = pickingProxy.component;
                    console.log(atom);
                    console.log(component);
                    document.getElementById('{molname_id}').innerText = `${{component.name}}`;
                    document.getElementById('{atom_id}').innerText = `${{atom.index}}`;
                    }}
            }});
        ''')


import unittest
import subprocess

from fragmenstein.cli import FragmensteinParser
import pkg_resources
from fragmenstein.demo import TestSet
import fragmenstein.demo.test_mols as mol_folder
from pathlib import Path

class CliTests(unittest.TestCase):

    def command(self, command):
        process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, err = process.communicate()
        return out, err, process.returncode

    def test_monster_combine(self):
        parser = FragmensteinParser()
        return parser(['monster', 'combine',
                '-i', str(Path(mol_folder.__file__).parent / 'mac-x0138.mol'),
                      str(Path(mol_folder.__file__).parent / 'mac-x0398.mol')
                ])

    def test_monster_place(self):
        parser = FragmensteinParser()
        return parser(['monster', 'place',
                '-i', str(Path(mol_folder.__file__).parent / 'mac-x0138.mol'),
                      str(Path(mol_folder.__file__).parent / 'mac-x0398.mol'),
                '-s', 'COCCON'
                ])








from ..victor import Victor
from .mcs_monster import MCSMerger
from rdkit import Chem

class Mictor(Victor):
    # MCS hack

    def _calculate_combination_chem(self):
        """
        The rdkit part. Monster is used to combine the hits.
        """
        print("modded `_calculate_combination_chem` for `MCSMerger` called")
        self._harmonize_warhead_combine()
        # TODO Does combine not need attachment??
        self.monster = MCSMerger(self.hits)
        self.monster.combine(embed_mode=0)
        # self.monster_throw_on_discard is not applicable
        self.mol = self.monster.positioned_mol
        self.mol.SetProp('_Name', self.monster.merged_name)
        self.smiles = Chem.MolToSmiles(self.mol)
        # making folder.
        self.make_output_folder()
        self.unmatched = []



########################################################################################################################

__doc__ = \
    """
These are extra functionality for Igor
    """

########################################################################################################################

import requests, shutil
from typing import Optional, Dict, Union
from .pyrosetta_import import pyrosetta  # the real mcCoy or a mock.
from ._igor_base import _IgorBase
from rdkit import Chem
from rdkit.Chem import AllChem
from rdkit.Chem.Draw import SimilarityMaps
from enum import Enum


class _IgorUtils(_IgorBase):

    def dock(self) -> pyrosetta.Pose:
        """
        Docks the pose the normal way and without constraints.

        :return:
        """
        docked = self.pose.clone()
        # self.ligand_residue[0] is pose index. This is a bit weird / hacky for now:
        # the pdb_info needs resetting everything chain A.
        docked.pdb_info().set_resinfo(res=self.ligand_residue[0], chain_id='B', pdb_res=1)
        docked.remove_constraints()
        pyrosetta.rosetta.protocols.docking.setup_foldtree(docked, 'A_B', pyrosetta.Vector1([1]))
        scorefxn = pyrosetta.create_score_function('ligand')
        docking = pyrosetta.rosetta.protocols.docking.DockMCMProtocol()
        docking.set_scorefxn(scorefxn)
        docking.apply(docked)
        return docked

    def per_atom_scores(self,
                        pose: Optional[pyrosetta.Pose]=None,
                        target_res: Optional[int]=None,
                        scorefxn: Optional[pyrosetta.ScoreFunction]=None) -> Dict[str, Dict[str, float]]:
        """
        Per atom scores are generally a bad idea as a score relative to something else is better.
        So do treat with the appropriate caution.
        NB. these scores will not sum to the that of the residue.

        Given a pose, a target_res and a scorefxn return a dict of per atom scores:

        * Lenard-Jones attraction 6-term
        * Lenard-Jones repulsion 12-term
        * Solvatation (zero)
        * Electrostatic interactions

        :param pose:
        :param target_res:
        :param scorefxn:
        :return: a dict of atom names to dict of 'lj_atr', 'lj_rep', 'fa_solv', 'fa_elec' to value
        """
        # Defaults
        if target_res is None:
            target_res = self.ligand_residue[0]
        if pose is None:
            pose = self.pose
        if scorefxn is None:
            # The minimisation is done under cartesian setting though.
            # and constrained!
            scorefxn = pyrosetta.get_fa_scorefxn()
        # Prep
        score_types = ['lj_atr', 'lj_rep', 'fa_solv', 'fa_elec']
        residue = pose.residue(target_res)
        scores = {residue.atom_name(i): {st: 0 for st in score_types} for i in range(1, residue.natoms() + 1)}
        # Iterate per target residue's atom per all other residues' atoms
        for i in range(1, residue.natoms() + 1):
            iname = residue.atom_name(i)
            for r in range(1, pose.total_residue() + 1):
                other = pose.residue(r)
                for o in range(1, other.natoms() + 1):
                    score = pyrosetta.toolbox.atom_pair_energy.etable_atom_pair_energies(residue,
                                                                                         i,
                                                                                         other,
                                                                                         o,
                                                                                         scorefxn)
                    for st, s in zip(score_types, score):
                        # technically only fa_intra_rep is 100 fold less than fa_rep
                        scores[iname][st] += s if target_res != r else s/100
        return scores

    def display_energy(self,
                       minimized_mol: Union[Chem.Mol, None],
                       term: Union[str, Enum]='all',
                       scorefxn: Optional[pyrosetta.ScoreFunction]=None,
                       colorMap:str='coolwarm') -> Dict[str, float]:
        """
        Given the ``minimized_mol`` display the energy contributions.
        As Victor has this while Igor is naive of bond orders, if None is passed it will made due with no bond order.

        ``term`` is the term to use:

        * ``all`` will be ``['lj_atr', 'lj_rep', 'fa_solv', 'fa_elec']``
        * ``lenard-jones`` will ``['lj_atr', 'lj_rep']``
        * while ``['lj_atr', 'lj_rep', 'fa_solv', 'fa_elec']`` will be the individual ones.

        As this is rather obscure, there is an enum as an attribute of ``display_energy`` method called ``Term``,
        with the values ``.display.Term.ALL``, ``.display.Term.ATTRACTION``, ``.display.Term.REPULSION``,
        ``.display.Term.LENARDJONES``, ``.display.Term.SOLVATATION``, ``.display.Term.ELECTROSTATIC``
        which is possibly more sane (e.g. ``igor.display_energy(minimized_mol, term=igor.display.Term.ELECTROSTATIC)``).
        """
        if minimized_mol is None:
            minimized_mol:Chem.Mol = self.mol_from_pose(add_dummy=False)
        if scorefxn is None:
            scorefxn:pyrosetta.ScoreFunction = pyrosetta.get_fa_scorefxn()
        # ['all', 'lenard-jones', 'lj_atr', 'lj_rep', 'fa_solv', 'fa_elec']
        # deal with choices of terms:
        if isinstance(term, Enum):
            term = term.value
        weights = {'lj_atr': scorefxn.get_weight(pyrosetta.rosetta.core.scoring.ScoreType.fa_atr),
                   'lj_rep': scorefxn.get_weight(pyrosetta.rosetta.core.scoring.ScoreType.fa_rep),
                   'fa_solv': scorefxn.get_weight(pyrosetta.rosetta.core.scoring.ScoreType.fa_sol),
                   'fa_elec': scorefxn.get_weight(pyrosetta.rosetta.core.scoring.ScoreType.fa_elec)}
        if term == 'all':
            pass
        elif term == 'lenard-jones':
            weights['fa_solv'] = 0
            weights['fa_elec'] = 0
        elif term in weights:
            for other in set(weights.keys()) - {term}:
                weights[other] = 0
        else:
            raise ValueError(f'Term {term} is not `all`,`lj_atr`,`lj_rep`,`fa_solv`,`fa_elec`')

        # get scores
        scores = self.per_atom_scores()

        def get_combined(atom: Chem.Atom):
            # combine the scores
            scored = scores[atom.GetPDBResidueInfo().GetName()]
            return sum([scored[k] * weights[k] for k in scored])

        contribs = [get_combined(atom) for atom in minimized_mol.GetAtoms()]
        # make into a 2D mol
        mol = Chem.Mol(minimized_mol)
        AllChem.Compute2DCoords(mol)
        # displays automatically:
        SimilarityMaps.GetSimilarityMapFromWeights(mol, contribs, colorMap=colorMap, contourLines=10)
        return {atom.GetProp('atom_name'): round(get_combined(atom), 1) for atom in minimized_mol.GetAtoms()}

    @classmethod
    def download_map(cls, pdbcode: str, filename: str):
        """
        Download to disk a CCP4 map of a given PDB code
        :param pdbcode:
        :param filename:
        :return:
        """
        assert '.ccp4' in filename, f'This downloads ccp4 maps ({filename})'
        r = requests.get(f'http://www.ebi.ac.uk/pdbe/coordinates/files/{pdbcode.lower()}.ccp4', stream=True)
        if r.status_code == 200:
            with open(filename, 'wb') as f:
                r.raw.decode_content = True
                shutil.copyfileobj(r.raw, f)

    @classmethod
    def relax_with_ED(cls, pose, ccp4_file: str, constraint_file: Optional[str] = None) -> None:
        """
        Relaxes ``pose`` based on the ccp4 electron density map provided. See ``download_map`` to download one.
        :param pose:
        :param ccp4_file: download map from ePDB
        :return: Relaxes pose in place
        """
        scorefxnED = pyrosetta.get_fa_scorefxn()
        ED = pyrosetta.rosetta.core.scoring.electron_density.getDensityMap(ccp4_file)
        sdsm = pyrosetta.rosetta.protocols.electron_density.SetupForDensityScoringMover()
        sdsm.apply(pose)
        ## Set ED constraint
        elec_dens_fast = pyrosetta.rosetta.core.scoring.ScoreType.elec_dens_fast
        scorefxnED.set_weight(elec_dens_fast, 30)
        ## Set generic constraints
        if constraint_file:
            stm = pyrosetta.rosetta.core.scoring.ScoreTypeManager()
            for contype_name in ("atom_pair_constraint", "angle_constraint", "dihedral_constraint"):
                contype = stm.score_type_from_name(contype_name)
                scorefxnED.set_weight(contype, 5)
            setup = pyrosetta.rosetta.protocols.constraint_movers.ConstraintSetMover()
            setup.constraint_file(constraint_file)
            setup.apply(pose)
        ## Relax
        for w in (30, 20, 10):
            scorefxnED.set_weight(elec_dens_fast, w)
            relax = pyrosetta.rosetta.protocols.relax.FastRelax(scorefxnED, 5)
            relax.apply(pose)

    @staticmethod
    def init_pyrosetta():
        """
        Common options for pyrosetta. Does not use `pyrosetta_help` atm
        """
        pyrosetta.init(extra_options='-no_optH false -ex1 -ex2 -mute all -ignore_unrecognized_res false ' +\
                                     '-load_PDB_components false -ignore_waters true')


class Term(Enum):
    ALL = 'all'
    ATTRACTION = 'lj_atr'
    REPULSION = 'lj_rep'
    LENARDJONES = 'lenard-jones'
    SOLVATATION = 'fa_solv'
    ELECTROSTATIC = 'fa_elec'

setattr(_IgorUtils.display_energy, 'Term', Term)


import pandas as pd
from rdkit.Chem import PandasTools

class LabExtras:

    error_classifications ={'SUCCESS': 'success',
                            'UNCLASSIFIED': 'unclassified',
                            'ConnectionError': 'distance',  # legacy
                            'DistanceError': 'distance',
                            'RectificationError': 'incorrect rectification',
                            'FullOverlapError': 'distance',
                            'TimeoutError': 'timeout',
                            'No sub-structure match': 'incorrect parameterisation #1',
                            'Violation occurred': 'incorrect rectification #1',
                            'KekulizeException': 'incorrect rectification #1',
                            'AtomValenceException': 'incorrect rectification #1',
                            'Bad Conformer Id': 'incorrect rectification #1',
                            ('rosetta', 'not found'): 'incorrect parameterisation #2',
                            ('rosetta', 'nan'): 'incorrect parameterisation #3',
                            'but not the minimised': 'incorrect parameterisation #4',
                            'UtilityExitException': 'incorrect parameterisation #5',
                            'TypeError': 'embedding error',
                            'RecursionError': 'embedding error',
                            'utterly impossible': 'embedding error'
                      }
    @classmethod
    def error_classify(cls, text):
        if not text:
            return cls.error_classifications['SUCCESS']
        text = str(text)
        for key, value in cls.error_classifications.items():
            if isinstance(key, tuple):
                if all(x in text for x in key):
                    return value
            elif key in text:
                return value
            else:
                return cls.error_classifications['UNCLASSIFIED']

    @classmethod
    def convert_to_sdf(self,
                       df: pd.DataFrame,
                       filename: str = f'fragmenstein.sdf',
                       acceptable_only = True,
                       sort_values='∆∆G',
                       name='name',
                       mol_name='minimized_mol'):
        if acceptable_only:
            df = df.loc[(df.outcome == 'acceptable')]
        short = df.sort_values(sort_values).reset_index().drop_duplicates(name)

        PandasTools.WriteSDF(df=short,
                             out=filename,
                             molColName=mol_name,
                             idName=name,
                             properties=['regarded', 'smiles', '∆∆G', '∆G_bound', '∆G_unbound',
                                         'comRMSD', 'N_constrained_atoms', 'N_unconstrained_atoms', 'runtime',
                                         'LE', 'outcome',
                                         'prcent_hybrid']
                             )



import argparse, logging
from ..victor import Victor
from typing import Optional, List


class FragmensteinParserBase:
    """
    See main module docstring for usage.
    """

    subcommand_help = '''Actions: monster, victor, laboratory, utils
    '''
    monster_help = '''monster'''
    laboratory_help = '''laboratory'''
    victor_help = '''victor'''
    utils_help = '''utils'''

    def __init__(self):
        self.parser = argparse.ArgumentParser(description=self.__doc__)
        subparsers = self.parser.add_subparsers(title='subcommands', help=self.subcommand_help)
        monster_parser = subparsers.add_parser('monster', help=self.monster_help)
        self._define_monster(monster_parser)
        laboratory_parser = subparsers.add_parser('laboratory', help=self.laboratory_help)
        self._define_laboratory(laboratory_parser)
        victor_parser = subparsers.add_parser('victor', help=self.victor_help)
        self._define_victor(victor_parser)
        utils_parser = subparsers.add_parser('utils', help=self.utils_help)
        self._define_utils(utils_parser)

    def __call__(self, cli_override: Optional[List[str]] = None):
        if cli_override:
            args = self.parser.parse_args(cli_override)
        else:
            args = self.parser.parse_args()
        args.func(args)

    def _define_laboratory(self, parser: argparse.ArgumentParser):
        raise NotImplementedError('virtual method')

    def _define_victor(self, parser: argparse.ArgumentParser):
        raise NotImplementedError('virtual method')

    def _define_monster(self, parser: argparse.ArgumentParser):
        raise NotImplementedError('virtual method')

    def _define_utils(self, parser: argparse.ArgumentParser):
        raise NotImplementedError('virtual method')


def set_verbose(count):
    log_map = {None: logging.FATAL,
               0: logging.CRITICAL,
               1: logging.ERROR,
               2: logging.WARNING,
               3: logging.INFO,
               4: logging.DEBUG}
    Victor.enable_stdout(log_map.get(count, logging.DEBUG))
    Victor.capture_logs()

def _add_common(parser: argparse.ArgumentParser, **settings):
    if settings.get('verbose', True):
        parser.add_argument('-v', '--verbose', action="count", help='verbose')
    if settings.get('hits', False):
        parser.add_argument('-i', '--hits', nargs='+', required=True, help='hit mol files')
    if settings.get('output', False):
        parser.add_argument('-o', '--output', default='.', help='output root folder')
    if settings.get('name', False):
        parser.add_argument('-n', '--name', default='fragmenstein', help='output name of molecule')
    if settings.get('template', False):
        parser.add_argument('-t', '--template', required=True, help='template PDB file')



from rdkit.Chem import rdFMCS
from ..error import FragmensteinError
from ..victor import Victor
from rdkit import Chem
from rdkit.Chem import AllChem


class Quicktor(Victor):
    """
    This class is intended to speed up the Victor placement step, by:

    * doing a strict MCS only
    * not allowing constrained atoms to move during RDKit minimisation and killing them if they fail.
    * doing a quick pyrosetta minimisation

    Example usage:

    .. code-block:: python
       lab = Laboratory(hits, pdb_block=pdb_block)
       lab.Victor = Quicktor
       to_be_placed: pd.DataFrame = place_input_validator(to_be_placed)
       placed: pd.DataFrame = lab.place(to_be_placed_df, n_cores=55, expand_isomers=False, timeout=60)
    """
    quick_reanimation = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.monster.matching_modes = [{'atomCompare': rdFMCS.AtomCompare.CompareElements,
                                        'bondCompare': rdFMCS.BondCompare.CompareOrder,
                                        'ringCompare': rdFMCS.RingCompare.PermissiveRingFusion,
                                        'ringMatchesRingOnly': True}]
        self.monster_mmff_minisation = True  # as is default. Repeated for emphasis.


    def _get_preminimized_undummied_monster(self) -> Chem.Mol:
        """
        This method is called by the plonking into structure methods.
        Not "positioning" as intended by ``monster`` is done.
        Opening a PDB in RDKit is doable but gets exponentially slow with chain length
        """
        mol = Chem.Mol(self.monster.positioned_mol)
        if self.monster_mmff_minisation:
            self.journal.debug(f'{self.long_name} - pre-minimising monster (MMFF)')
            if not self.monster.mmff_minimize(mol, allow_lax=False, ff_dist_thr=float('nan')):
                raise FragmensteinError('Could not preminize')
            ddG = self.monster.MMFF_score(mol, delta=True)
            if ddG > 100:
                raise FragmensteinError('Poor preminization')
        return AllChem.DeleteSubstructs(mol, Chem.MolFromSmiles('*'))



from .base import FragmensteinParserBase
from .monster import FragmensteinParserMonster
from .victor import FragmensteinParserVictor
from .laboratory import FragmensteinParserLaboratory
from .utils import FragmensteinParserUtils


class FragmensteinParser(FragmensteinParserMonster,
                         FragmensteinParserVictor,
                         FragmensteinParserLaboratory,
                         FragmensteinParserUtils,
                         FragmensteinParserBase):
    pass



__all__ = ['TestSet', 'MPro', 'BaseMolDataset']



from ._base_dataset_holder import BaseMolDataset, extend_doc
from rdkit import Chem
from typing import List
import os


# ----------------------------------------------------------------------------------------------------------------------
from . import test_mols

class TestSet(BaseMolDataset):
    """
    Test molecules for the tests.
    """
    dataset_package = test_mols

    @classmethod
    def get_template(cls):
        cls.raise_not_implemented()

    @classmethod
    def get_5SB7_mols(cls) -> List[Chem.Mol]:
        """
        In https://onlinelibrary.wiley.com/doi/10.1002/anie.202204052
        There is a lovely test case.
        Namely the followup bold4 is a merger of F36 and F04 hits.
        However... F04 (hello Hantzsch thiazole synthesis product) presents
        the problematic case that the thiazole can be flipped either way
        upon investigation of the crystal map
        and the followup bold4 is indeed a flipped merger.
        Therefore, without a custom_map the followup is incorrect.
        """
        return cls.get_sdf_mols('5SB7_mergers.sdf')

# ----------------------------------------------------------------------------------------------------------------------
from . import mpro_mols

class MPro(BaseMolDataset):
    """
    List of XChem hits of MPro from Fragalysis in Feb 2021.
    """

    dataset_package = mpro_mols

    @classmethod
    def file2mol_func(cls, filename: str) -> str:
        return os.path.splitext(filename)[0].replace('Mpro-', '').strip()

    @classmethod
    def mol2file_func(cls, hit_name: str) -> str:
        return f'Mpro-{hit_name}.mol'


# ----------------------------------------------------------------------------------------------------------------------
from . import mac1_mols

class Mac1(BaseMolDataset):
    """
    These are the fragments from https://pubmed.ncbi.nlm.nih.gov/33853786/
    binding he Macrodomain (Mac1 or NSP13) from SAR-COV-2
    """
    dataset_package = mac1_mols

extend_doc(TestSet)
extend_doc(MPro)
extend_doc(Mac1)






import argparse
import pandas as pd
from .base import FragmensteinParserBase, _add_common, set_verbose
from rdkit import Chem
from rdkit.Chem import PandasTools
from ..laboratory import Laboratory
from ..igor import Igor

class FragmensteinParserLaboratory:

    def _define_laboratory(self, parser: argparse.ArgumentParser):
        """fragmenstein laboratory combine -i hits.sdf -o out.sdf
           fragmenstein laboratory place -i hits.sdf -t table.csv -o out.sdf"""
        subparsers = parser.add_subparsers(title='operation', help='combine or place')
        combine_parser = subparsers.add_parser('combine', help='combine')
        _add_common(combine_parser, hits=False, output=True, template=True)
        parser.add_argument('-i', '--input', required=True, help='input sdf file')
        parser.add_argument('-d', '--table', default='output.csv', help='table output file')
        parser.add_argument('-s', '--sdf_outfile', default='output.sdf', help='sdf output file')
        parser.add_argument('-c', '--cores', default=1, type=int, help='number of cores to use')
        combine_parser.set_defaults(func=self.lab_combine)
        place_parser = subparsers.add_parser('place', help='place')
        place_parser.set_defaults(func=self.lab_place)

    def lab_combine(self, args: argparse.Namespace) -> str:
        set_verbose(args.verbose)
        Igor.init_pyrosetta()
        lab = Laboratory(pdbblock=open(args.template).read(), )
        with Chem.SDMolSupplier(args.input) as suppl:
            mols = list(suppl)
        combos: pd.DataFrame = lab.combine(mols, n_cores=args.cores)
        props = ['smiles', 'error','∆∆G', '∆G_bound', '∆G_unbound', 'comRMSD',
       'N_constrained_atoms', 'N_unconstrained_atoms', 'runtime', 'regarded',
       'disregarded', 'LE', 'outcome',
       'percent_hybrid']
        combos[props].to_csv(args.table)
        PandasTools.WriteSDF(combos, args.sdf_outfile,
                             'minimized_mol',
                             properties=props)

    def lab_place(self, args: argparse.Namespace) -> str:
        set_verbose(args.verbose)
        Igor.init_pyrosetta()
        lab = Laboratory(pdbblock=open(args.template).read(), )
        raise NotImplementedError('MF has not thought hard enough how this could be done')



import logging
import tempfile
import unittest, os
# ======================================================================================================================
from multiprocessing import Process

import pyrosetta

# ======================================================================================================================
from rdkit import Chem
from rdkit.Chem import AllChem

from fragmenstein import Monster, Victor, Igor, mpro_data, Walton
from fragmenstein.mpro import MProVictor
from typing import *
import numpy as np

from fragmenstein.demo import TestSet

# import logging
# Victor.enable_stdout(logging.DEBUG)
# ======================================================================================================================

class MonsterCombineTests(unittest.TestCase):
    def test_phenylene(self):
        # make carboxy and amide benzenes that overlap so that the end result is a phenylene where one ring is oxazine
        conjoined = Chem.MolFromSmiles('c3c1cccc2\C(=O)O/C(-N)c(c12)cc3')
        conjoined = next(AllChem.EnumerateStereoisomers(conjoined))
        before = Chem.MolToSmiles(conjoined)  # structure from wiki is not canonical
        AllChem.EmbedMolecule(conjoined)
        bonds = [conjoined.GetBondBetweenAtoms(0, 1).GetIdx(),
                 conjoined.GetBondBetweenAtoms(12, 11).GetIdx(),
                 conjoined.GetBondBetweenAtoms(8, 9).GetIdx()]
        fragged = Chem.FragmentOnBonds(conjoined, bonds, addDummies=False)
        fore = Chem.GetMolFrags(fragged, asMols=True, sanitizeFrags=False)[1]
        Chem.SanitizeMol(fore)
        bonds = [conjoined.GetBondBetweenAtoms(2, 1).GetIdx(),
                 conjoined.GetBondBetweenAtoms(12, 5).GetIdx(),
                 conjoined.GetBondBetweenAtoms(8, 6).GetIdx()]
        fragged = Chem.FragmentOnBonds(conjoined, bonds, addDummies=False)
        aft = Chem.GetMolFrags(fragged, asMols=True, sanitizeFrags=False)[0]
        Chem.SanitizeMol(aft)
        # merge them
        mol = Monster([fore, aft]).combine().positioned_mol
        after = Chem.MolToSmiles(Chem.RemoveHs(mol))
        self.assertEqual(before, after)

    def test_orthomethyltoluene(self):
        name = 'orthomethyltoluene'
        after = 'Cc1cccc(C)c1'
        toluene = TestSet.get_mol('toluene')
        rototoluene = TestSet.get_mol('rototoluene')
        mol = Monster(hits=[toluene, rototoluene]).combine(keep_all=True).positioned_mol
        gotten = Chem.MolToSmiles(Chem.RemoveHs(mol))
        self.assertEqual(gotten, after, f'{name} failed {gotten} (expected {after})')

    def test_peridimethylnaphthalene(self):
        name = 'peridimethylnaphthalene'
        after = 'Cc1cccc2cccc(C)c12'
        toluene = TestSet.get_mol('toluene')
        transtolueneF = TestSet.get_mol('transtoluene')
        transtolueneF.SetProp('_Name', 'transtoluene-fuse')
        mol = Monster(hits=[toluene, transtolueneF]).combine(keep_all=True).positioned_mol
        gotten = Chem.MolToSmiles(Chem.RemoveHs(mol))
        self.assertEqual(gotten, after, f'{name} failed {gotten} (expected {after})')

    def test_spirodituluene(self):
        name = 'spirodituluene'
        after = ('CC1CCC2(CCCC(C)C2)CC1',
                 'CC1C=CC2(C=C1)CC=CC(C)C2',
                 'C[C@@H]1C=CC[C@]2(C=C[C@H](C)C=C2)C1',
                 'C[C@@H]1C=CC[C@]2(C=C[C@@H](C)C=C2)C1',
                 'C[C@H]1C=CC[C@]2(C=C[C@H](C)C=C2)C1',
                 'C[C@H]1C=CC[C@]2(C=C[C@@H](C)C=C2)C1')
        toluene = TestSet.get_mol('toluene')
        transtolueneS = TestSet.get_mol('transtoluene2')
        transtolueneS.SetProp('_Name', 'transtoluene-spiro')
        # cmd.rotate('z', -90, 'rototoluene', camera=0)  # this predates capt. Robert Walton
        mol = Monster(hits=[toluene, transtolueneS]).combine(keep_all=True).positioned_mol
        gotten = Chem.MolToSmiles(Chem.RemoveAllHs(mol))
        self.assertIn(gotten, after, f'{name} failed {gotten} (expected {after})')

    def test_real_merger(self):
        # Victor.enable_stdout(logging.DEBUG)
        x0138 = TestSet.get_mol('mac-x0138')
        x0398 = TestSet.get_mol('mac-x0398')
        monster = Monster([x0398, x0138])
        monster.combine()
        self.assertEqual('Nc1nc2c3c(c(O)cc(N)c3n1)C(O)=N2',
                         Chem.MolToSmiles(Chem.RemoveHs(monster.positioned_mol)),
                         )



from ._victor_igor import _VictorIgor
from .minimalPDB import MinimalPDBParser
import os, re
from typing import *
from rdkit import Chem
from rdkit.Chem import AllChem
from rdkit_to_params import Constraints

# ----------------------------------------------------------------------------------------------------------------------

class _VictorCommon(_VictorIgor):

    def make_output_folder(self):
        os.makedirs(self.work_path, exist_ok=True)
        path = os.path.join(self.work_path, self.long_name)
        if not os.path.exists(path):
            os.mkdir(path)
        else:
            self.journal.warning(f'{self.long_name} - Folder {path} exists.')

    def _save_prerequisites(self):
        self._log_warnings()
        #  saving params
        self.journal.debug(f'{self.long_name} - saving params')
        params_file = os.path.join(self.work_path, self.long_name, self.long_name + '.params')
        self.params.dump(params_file)
        # saving holo
        self.journal.debug(f'{self.long_name} - saving holo (unminised)')
        holo_file = os.path.join(self.work_path, self.long_name, self.long_name + '.holo_unminimised.pdb')
        with open(holo_file, 'w') as w:
            w.write(self.unminimized_pdbblock)
        # saving constraint
        if self.constraint is not None:
            self.journal.debug(f'{self.long_name} - saving constraint')
            constraint_file = os.path.join(self.work_path, self.long_name, self.long_name + '.con')
            self.constraint.dump(constraint_file)
        else:  # basically impossible.
            constraint_file = ''
        return params_file, holo_file, constraint_file

    # =================== Constraint & attachment ======================================================================

    def _get_constraint(self, extra_constraint: Optional[str] = None) -> Union[Constraints, None]:
        # deal with covalent and non covalent separately
        if self.is_covalent:
            self.journal.debug(f'{self.long_name} - is covalent.')
            constraint = self._fix_covalent()
            if extra_constraint:
                constraint.custom_constraint += self.extra_constraint
            return constraint
        else:
            self.journal.debug(f'{self.long_name} - is not covalent.')
            constraint = self._fix_uncovalent()
            if extra_constraint:
                constraint.custom_constraint += self.extra_constraint
            return constraint

    def _fix_uncovalent(self):
        return Constraints.mock()

    def _fix_covalent(self):
        self.journal.debug(f'{self.long_name} - fixing for covalent')
        # to make life easier for analysis, CX is the attachment atom, CY is the one before it.
        war_def = self._get_war_def()
        warhead = Chem.MolFromSmiles(war_def['covalent'])
        self.params.rename_by_substructure(warhead, war_def['covalent_atomnames'])
        cov_def = [d for d in self.covalent_definitions if d['residue'] == self.covalent_resn][0]
        self.journal.debug(f'{self.long_name} - has a {war_def["name"]}')
        cons = Constraints(smiles=(war_def['covalent'], cov_def['smiles']),
                           names=[*war_def['covalent_atomnames'], *cov_def['atomnames']],
                           ligand_res=self.ligand_resi,
                           target_res=self.covalent_resi)
        # user added constraint
        if 'constraint' in war_def:
            cons.custom_constraint = war_def['constraint']
        return cons

    def add_extra_constraint(self, new_constraint:Union[str]=None):
        if new_constraint is None:
            return # do nothing
        new_constraint = new_constraint.strip()
        if self.extra_constraint is None:
            self.extra_constraint = new_constraint
        self.extra_constraint = self.extra_constraint.strip() + '\n' + new_constraint.strip()

    def make_coordinate_constraints_for_placement(self,
                                    mol: Optional[Chem.Mol] = None,
                                    origins: Optional[List[List[str]]] = None,
                                    std: Optional[List[float]] = None,
                                    mx: Optional[List[float]] = None) -> str:
        """
        See also ``make_coordinate_constraints_for_combination`` in combine.
        This is the normal function and uses the origin data,
        while the other constrains based on lack of novel attribute.

        :param mol: self.monster.positioned_mol if ommitted
        :param origins: self.monster.origin_from_mol(self.monster.positioned_mol)  if ommitted
            list of list of names of hit atoms used as original position
        :param std: self.monster.stdev_from_mol(mol)(self.monster.positioned_mol)  if omitted
            list of standard devs
        :param mx: elf.monster.max_from_mol(self.monster.positioned_mol)  if omitted
            list of maximum euclidean distance
        :return:
        """
        lines = []
        if mol is None:
            mol = self.monster.positioned_mol
        if origins is None:
            origins = self.monster.origin_from_mol(mol)
        if std is None:
            std = self.monster.stdev_from_mol(mol)
        if mx is None:
            mx = self.monster.max_from_mol(mol)
        conf = self.monster.positioned_mol.GetConformer()
        # Calculate
        for i in range(mol.GetNumAtoms()):
            if len(origins[i]) > 0:
                atom: Chem.Atom = mol.GetAtomWithIdx(i)
                if atom.GetAtomicNum() < 2:  # noqa
                    # zahl of 0 is *, 1 is H
                    continue
                elif atom.GetPDBResidueInfo() is None:
                    self.journal.critical(f'Atom {i} ({atom.GetSymbol()}) has no name!')
                    continue
                pos = conf.GetAtomPosition(i)
                if self.constraint_function_type.upper() == 'HARMONIC':
                    fxn = f'HARMONIC 0 {std[i] + 1}'
                elif self.constraint_function_type.upper() == 'FLAT_HARMONIC':
                    if len(origins[i]) > 1:
                        fxn = f'FLAT_HARMONIC 0 1.0 {mx[i]}'
                    else:
                        fxn = f'HARMONIC 0 1.0'
                elif self.constraint_function_type.upper() == 'BOUNDED':
                    fxn = f'BOUNDED 0 {mx[i]} 1 0.5 TAG'
                else:
                    raise ValueError(f'{self.constraint_function_type} is not HARMONIC or FADE or BOUNDED')
                atomname = atom.GetPDBResidueInfo().GetName()
                line = f'CoordinateConstraint {atomname} {self.ligand_resi} ' + \
                             f'CA {self.covalent_resi} ' + \
                             f'{pos.x} {pos.y} {pos.z} {fxn}\n'
                if 'nan' in line:
                    # todo: This is serious.
                    self.journal.warning(f'{atomname} lacks coordinates')
                    continue
                lines.append(line)
        return ''.join(lines)

    def make_coordinate_constraints_for_combination(self):
        """
        See also ``cls.make_coordinate_constraints_for_placement``.
        This operates based on ``atom.HasProp('_Novel')``, not origins!
        :return:
        """
        lines = []
        conf: Chem.Conformer = self.monster.positioned_mol.GetConformer()
        atom: Chem.Atom
        for i, atom in enumerate(self.monster.positioned_mol.GetAtoms()):
            if atom.GetAtomicNum() < 2:  # noqa
                # zahl of 0 is *, 1 is H
                continue
            elif atom.HasProp('_Novel') and atom.GetBoolProp('_Novel'):
                continue  # novels
            elif atom.GetPDBResidueInfo() is None:
                self.journal.critical(f'Atom {i} ({atom.GetSymbol()}) has no name!')
                continue
            else:
                pos = conf.GetAtomPosition(i)
                fxn = f'HARMONIC 0 1'  # the other do not make sense here.

                line = f'CoordinateConstraint {atom.GetPDBResidueInfo().GetName()} {self.ligand_resi} ' + \
                             f'CA {self.covalent_resi} ' + \
                             f'{pos.x} {pos.y} {pos.z} {fxn}\n'
                if 'nan' in line:
                    # todo: This is serious.
                    self.journal.warning(f'Atom {i} lacks coordinates')
                    continue
                lines.append(line)
        return ''.join(lines)

    # ------------------------------------------------------------------------------------------------------------------

    def _get_attachment_from_pdbblock_via_pymol(self) -> Union[None, Chem.Mol]:
        """
        DEPRACATED. FORMER CODE FOR ``_get_attachment_from_pdbblock``

        Yes, yes, I see the madness in using pymol to get an atom for rdkit to make a pose for pyrosetta.
        Hence why `find_attachment` will replace it.
        """
        import pymol2

        self.journal.debug(f'{self.long_name} - getting attachemnt atom')
        if not self.covalent_resn:
            return None
        else:
            if isinstance(self.covalent_resi, str):
                resi, chain = re.match('(\d+)(\w)', self.covalent_resi).groups()
                resi = int(resi)
            else:
                resi = self.covalent_resi
                chain = None
            with pymol2.PyMOL() as pymol:
                pymol.cmd.read_pdbstr(self.apo_pdbblock, 'prot')
                if self.covalent_resn == 'CYS':
                    name = 'SG'
                else:
                    raise NotImplementedError('only done for cys atm')
                try:
                    if chain is not None:
                        pdb = pymol.cmd.get_pdbstr(f'resi {resi} and name {name} and chain {chain}')
                    else:
                        pdb = pymol.cmd.get_pdbstr(f'resi {resi} and name {name}')
                except KeyboardInterrupt as err:
                    raise err
                except:
                    pdb = pymol.cmd.get_pdbstr(f'resi {resi} and name {name}')
                return Chem.MolFromPDBBlock(pdb)

    def _get_attachment_from_pdbblock(self) -> Union[None, Chem.Mol]:
        """
        ``find_attachment`` finds the actual attachment atom not the stated one.
        The historic code for this is now ``_get_attachment_from_pdbblock_via_pymol``
        :return:
        """
        if not self.covalent_resn:
            return None
        if self.covalent_resn == 'CYS':
            name = 'SG'
        else:
            raise NotImplementedError('only done for cys atm')
        if isinstance(self.covalent_resi, str):
            resi, chain = re.match('(\d+)(\w)', self.covalent_resi).groups()
            resi = int(resi)
        else:
            resi = self.covalent_resi
            chain = None
        parser = MinimalPDBParser(self.apo_pdbblock)
        for atom_row in parser.coordinates:
            if parser.get_residue_index(atom_row) != resi:
                continue
            if chain is not None and parser.get_chain(atom_row) != chain:
                continue
            if parser.get_atomname(atom_row) == name.strip():
                return Chem.MolFromPDBBlock(atom_row)
        return None

    def _get_war_def(self):
        for war_def in self.warhead_definitions:
            warhead = Chem.MolFromSmiles(war_def['covalent'])
            if self.mol.HasSubstructMatch(warhead):
                return war_def
        else:
            if self.mol.HasSubstructMatch(Chem.MolFromSmiles('*C')):
                self.journal.warning('Unknown type of covalent')
                return {'name': 'unknown',
                        'covalent': 'C~C*',
                        'covalent_atomnames': ['CY', 'CX', 'CONN1'],
                        'noncovalent': 'C~[C+]',  # clearly not
                        'noncovalent_atomnames': ['CY', 'CX']
                        }
            else:
                raise ValueError(f'{self.long_name} - Unsure what the warhead is {self.smiles}.')

    @classmethod
    def inventorize_warheads(cls, hits: List[Chem.Mol], covalent_form: bool = True) -> List[str]:
        """
        Get the warhead types of the list of hits

        :param hits:
        :param covalent_form: Are the hits already covalent (with *)
        :return: list of non-covalent, chloroacetimide, etc.
        """
        inventory = ['noncovalent'] * len(hits)
        for war_def in cls.warhead_definitions:
            wh = cls._get_warhead_from_definition(war_def, covalent_form)
            for i, hit in enumerate(hits):
                if hit.HasSubstructMatch(wh):
                    inventory[i] = war_def['name']
        return inventory

    @classmethod
    def _get_warhead_from_name(cls, warhead_name: str, covalent_form: bool) -> Chem.Mol:
        """
        get_warhead_definition returns a definition, this retursn a mol.

        :param warhead_name:
        :param covalent_form:
        :return:
        """
        war_def = cls.get_warhead_definition(warhead_name)
        wh = cls._get_warhead_from_definition(war_def, covalent_form)
        return wh

    @classmethod
    def _get_warhead_from_definition(cls, war_def: dict, covalent_form: bool):
        if covalent_form:
            wh = Chem.MolFromSmiles(war_def['covalent'])
        else:
            wh = Chem.MolFromSmiles(war_def['noncovalent'])
        return wh

    @classmethod
    def get_warhead_definition(cls, warhead_name: str):
        return cls._get_warhead_definitions(warhead_name)[0]

    @classmethod
    def _get_warhead_definitions(cls, warhead_name: str):
        """
        It is unlikely that alternative definitions are present. hence why hidden method.

        :param warhead_name:
        :return:
        """
        options = [wd for wd in cls.warhead_definitions if wd['name'] == warhead_name.lower()]
        if len(options) == 0:
            raise ValueError(f'{warhead_name} is not valid.')
        else:
            return options

    @classmethod
    def make_all_warhead_combinations(cls, smiles: str, warhead_name: str, canonical=True) -> Union[dict, None]:
        """
        Convert a unreacted warhead to a reacted one in the SMILES

        :param smiles: unreacted SMILES
        :param warhead_name: name in the definitions
        :param canonical: the SMILES canonical? (makes sense...)
        :return: dictionary of SMILES
        """
        mol = Chem.MolFromSmiles(smiles)
        war_def = cls.get_warhead_definition(warhead_name)
        ncv = Chem.MolFromSmiles(war_def['noncovalent'])
        if mol.HasSubstructMatch(ncv):
            combinations = {}
            for wd in cls.warhead_definitions:
                x = Chem.ReplaceSubstructs(mol, ncv, Chem.MolFromSmiles(wd['covalent']),
                                           replacementConnectionPoint=0)
                combinations[wd['name'] + '_covalent'] = Chem.MolToSmiles(x[0], canonical=canonical)
                x = Chem.ReplaceSubstructs(mol, ncv, Chem.MolFromSmiles(wd['noncovalent']),
                                           replacementConnectionPoint=0)
                combinations[wd['name'] + '_noncovalent'] = Chem.MolToSmiles(x[0], canonical=canonical)
            return combinations
        else:
            return None

    def harmonize_warheads(self, hits, warhead_harmonisation, covalent_form=True):
        """
        Harmonises and marks the atoms with `_Warhead` Prop.

        :param hits:
        :param warhead_harmonisation:
        :param covalent_form:
        :return:
        """
        inventory = self.inventorize_warheads(hits, covalent_form)
        # mark warhead atoms.
        for hit, warhead_name in zip(hits, inventory):
            if warhead_name != 'noncovalent':
                wh = self._get_warhead_from_name(warhead_name, covalent_form)
                match = hit.GetSubstructMatch(wh)
                if match == ():
                    self.journal.warning(f'{self.long_name} - failed to mark warhead. What is it??')
                else:
                    for i in match:
                        hit.GetAtomWithIdx(i).SetBoolProp('_Warhead', True)
        # harmonise
        if warhead_harmonisation == 'keep':
            return hits
        elif warhead_harmonisation == 'strip':
            new_hits = []
            for hit, warhead_name in zip(hits, inventory):
                if warhead_name != 'noncovalent':
                    wh = self._get_warhead_from_name(warhead_name, covalent_form)
                    nhit = AllChem.DeleteSubstructs(hit, wh)
                    Chem.SanitizeMol(nhit)
                    new_hits.append(nhit)
                else:
                    new_hits.append(hit)

            return new_hits
        elif warhead_harmonisation == 'first':
            if len(set(inventory) - {'noncovalent'}) <= 1:
                return hits
            else:
                first = None
                new_hits = []
                for hit, warhead_name in zip(hits, inventory):
                    if warhead_name != 'noncovalent':
                        if first is None:
                            first = warhead_name
                            new_hits.append(hit)
                        elif warhead_name == first:
                            new_hits.append(hit)
                        else:
                            wh = self._get_warhead_from_name(warhead_name, covalent_form)
                            nhit = AllChem.DeleteSubstructs(hit, wh)
                            Chem.SanitizeMol(nhit)
                            new_hits.append(nhit)
                    else:
                        new_hits.append(hit)
                return new_hits
        else: # it is a warhead name.
            raise NotImplementedError

    def MMFF_score(self, mol: Optional[Chem.Mol] = None, delta: bool = False) -> float:
        if mol is None:
            mol = self.igor.mol_from_pose()
        return self.monster.MMFF_score(mol, delta=delta)



import os
import tempfile
import unittest

# ======================================================================================================================
from rdkit import Chem

from fragmenstein import Victor, Igor
from fragmenstein.demo import TestSet

# ======================================================================================================================

class MultivictorPlaceTests(unittest.TestCase):

    def setUp(self):
        Igor.init_pyrosetta()

    def test_multivictor(self):
        from fragmenstein import MultiVictorPlacement
        to_place = TestSet.get_mol('placed_example1')
        pdb_block = TestSet.get_text('apo_example1.pdb')
        smiles = Chem.MolToSmiles(to_place)
        hits = [TestSet.get_mol('x0032_0A'),
                #TestSet.get_mol('x0103_0A')
                ]
        # Victor.enable_stdout(level=logging.ERROR)
        with tempfile.TemporaryDirectory() as tmpdir:
            Victor.work_path = os.path.join(tmpdir, "multivictor_out")
            mv = MultiVictorPlacement(hits=hits, pdb_block=pdb_block)
            mv.place(smiles, number_runs=4)
            # print(mv.retrieve_best_victor())
            # print(mv.retrieve_scores())
            self.assertLess(mv.retrieve_scores()[0], -7)


if __name__ == '__main__':
    unittest.main()



########################################################################################################################

__doc__ = \
    """
Place followup
    """

########################################################################################################################

from typing import Optional, Dict, List
from warnings import warn

from rdkit import Chem

from ._place_modes import _MonsterBlend
from .mcs_mapping import IndexMap

class _MonsterPlace(_MonsterBlend):
    """
    * _MonsterBase
    * _MonsterCommunal
    * _MonsterTracker
    * _MonsterMerge
    * _MonsterCombine
    * _MonsterMap
    * _MonsterChimera
    * _MonsterRefine
    * (_MonsterFull, _MonsterPartial, _MonsterNone) >
    * _MonsterBlend
    * _MonsterPlace
    """

    def place(self,
              mol: Chem.Mol,
              attachment: Optional[Chem.Mol] = None,
              custom_map: Optional[Dict[str, Dict[int, int]]] = None,
              merging_mode: str = 'expansion',
              enforce_warhead_mapping:bool=True):
        """
        Positioned a given mol based on the hits. (Main entrypoint)
        accepts the argument `merging_mode`, by default it is "expansion",
        but was "permissive_none",
        which call `.by_expansion` and `.no_blending(broad=True)` respectively.
        "off" (does nothing except fill the attribute ``initial_mol``),
        "full" (`.full_blending()`),
        "partial" (`.partial_blending()`)
        and "none" (`.no_blending()`, but less thorough)
        are accepted.

        :param mol:
        :param attachment: This the SG of the cysteine if covalent
        :param custom_map: Dict of hit_name to Dict of hit_idx to followup_idx
        :param merging_mode:
        :return:
        """
        if not mol.HasProp('_Name'):
            mol.SetProp('_Name', 'followup')
        assert not any([mol.GetProp('_Name') == h.GetProp('_Name') for h in self.hits]), 'Placement has the same name as a hit!'
        self.initial_mol, self.attachment = self._parse_mol_for_place(mol, attachment)
        custom_map = self.fix_custom_map(custom_map)
        if enforce_warhead_mapping and self.attachment:
            self.journal.debug('Enforcing warhead mapping')
            custom_map: Dict[str, Dict[int, int]] = self._add_warhead_mapping(custom_map)
        self.custom_map: Dict[str, Dict[int, int]] = custom_map
        # do the mol names match?
        self._validate_custom_map()
        # Reset
        self.unmatched = []
        self.mol_options = []
        # do calculations
        self.merging_mode = merging_mode
        if merging_mode == 'off':
            pass
        elif merging_mode == 'full':
            self.full_blending()
        elif merging_mode == 'partial':
            self.partial_blending()
        elif merging_mode == 'none_permissive' or merging_mode == 'permissive_none':
            self.no_blending(broad=True)
        elif merging_mode == 'none':
            self.no_blending()
        elif merging_mode == 'expansion':
            self.by_expansion()
        else:
            valid_modes = ('full', 'partial', 'none', 'none_permissive', 'off', 'expansion')
            raise ValueError(
                f"Merging mode can only be {'| '.join(valid_modes)}, not '{merging_mode}'")
        return self

    def place_smiles(self,
                     smiles: str,
                     long_name: Optional[str]=None,
                     **kwargs):
        mol = Chem.MolFromSmiles(smiles)
        mol.SetProp('_Name', long_name if long_name else 'followup')
        self.place(mol=mol, **kwargs)
        return self

    def _parse_mol_for_place(self,
                             mol: Chem.Mol,
                             attachment: Optional[Chem.Mol] = None):
        """
        Atachment is a single atom molecule, say the SG atom of a CYS, onto which the mol is placed.
        :param mol:
        :param attachment:
        :return:
        """
        # ------------- store mol ---------------------------------------
        if mol.HasSubstructMatch(self.dummy) and attachment:
            pass
        elif mol.HasSubstructMatch(self.dummy):
            warn('No attachment atom provided but dummy atom present --- ignoring.')
            attachment = None
        elif attachment:
            warn('Attachment atom provided but dummy atom not present --- ignoring.')
            attachment = None
        else:
            attachment = None
        return mol, attachment

    def _validate_custom_map(self):
        """
        What on earth has the user submitted as custom_map?

        :return:
        """
        # add missing keys
        hit_names:List[str] = []
        for hit in self.hits:  #: Chem.Mol
            hit_name:str = hit.GetProp('_Name')
            hit_names.append(hit_name)
            if hit_name not in self.custom_map:
                self.custom_map[hit_name] = {}
        # check for incorrect keys
        for hit_name in self.custom_map:
            if hit_name not in hit_names:
                raise ValueError(f"Custom map contains key '{hit_name}' which is not in hits ({hit_names}).")

    def _add_warhead_mapping(self, custom_map: Dict[str, IndexMap]) -> Dict[str, IndexMap]:
        """
        Add the warhead mapping to the custom_map.
        Does not use the warhead definition as it may be missing.

        :param custom_map:
        :return:
        """
        q: Chem.rdchem.QueryAtom = Chem.rdqueries.AtomNumEqualsQueryAtom(0)
        dummies = list(self.initial_mol.GetAtomsMatchingQuery(q))
        if len(dummies) == 0:
            self.journal.warning('The provided molecule is not covalent...')
            return custom_map
        if len(dummies) > 1:
            self.journal.info('More than one dummy atom found: cannot enforce warhead mapping.')
            return custom_map
        followup_dummy = dummies[0]

        def get_neighs(atom: Chem.Atom, old: Chem.Atom) -> List[Chem.Atom]:
            """get hetero neighs sorted by weight, except for the previous one"""
            neighs = [n for n in atom.GetNeighbors() if n.GetAtomicNum() != 1 and n.GetIdx() != old.GetIdx()]
            return sorted(neighs, key=lambda n: n.GetAtomicNum())

        for hit in self.hits:
            dummies = list(hit.GetAtomsMatchingQuery(q))
            if len(dummies) != 1:
                continue
            hit_dummy = dummies[0]
            hit_name = hit.GetProp('_Name')
            if hit_name not in custom_map:
                custom_map[hit_name] = {}
            # the hit and followup have a dummy atom:
            custom_map[hit_name][hit_dummy.GetIdx()] = followup_dummy.GetIdx()
            followup_atom = followup_dummy
            hit_atom = hit_dummy
            followup_old, hit_old = followup_atom, hit_atom
            while True:
                # crawl up the network
                followup_hetero_neighs = get_neighs(followup_atom, followup_old)
                hit_hetero_neighs = get_neighs(hit_atom, hit_old)
                if len(followup_hetero_neighs) == 1 and len(hit_hetero_neighs) == 1:
                    followup_old, hit_old = followup_atom, hit_atom
                    followup_atom = followup_hetero_neighs[0]
                    hit_atom = hit_hetero_neighs[0]
                    custom_map[hit_name][hit_atom.GetIdx()] = followup_atom.GetIdx()
                elif len(followup_hetero_neighs) == 2 and len(hit_hetero_neighs) == 2 and \
                        followup_hetero_neighs[0].GetSymbol() == 'C' and \
                        hit_hetero_neighs[0].GetSymbol() == 'C' and \
                        followup_hetero_neighs[1].GetSymbol() in ('O', 'N') and \
                        hit_hetero_neighs[1].GetSymbol() in ('O', 'N'):
                    custom_map[hit_name][hit_hetero_neighs[0].GetIdx()] = followup_hetero_neighs[0].GetIdx()
                    custom_map[hit_name][hit_hetero_neighs[1].GetIdx()] = followup_hetero_neighs[1].GetIdx()
                    followup_old, hit_old = followup_atom, hit_atom
                    followup_atom, hit_atom = followup_hetero_neighs[0], hit_hetero_neighs[0]
                else:
                    break
        return custom_map



import warnings

from rdkit.Chem import rdmolops

import itertools, operator
import json
from collections import Counter
from collections import defaultdict
from typing import Optional, Dict, List, Tuple, Set, Unpack, Union  # noqa cf. .legacy monkeypatch
from warnings import warn

import numpy as np
from rdkit import Chem
from rdkit.Chem import AllChem
from rdkit.Chem import rdFMCS, rdMolAlign, rdmolops
from rdkit.Chem import rdmolops
from rdkit.Geometry.rdGeometry import Point3D

from .._communal import _MonsterCommunal
from .._merge import _MonsterMerge
from ..unmerge_mapper import Unmerge
from ..mcs_mapping import SpecialCompareAtoms, IndexMap, ExtendedFMCSMode, transmute_FindMCS_parameters, flip_mapping
from ._refine import _MonsterRefine
from ...error import FragmensteinError

class _MonsterPartial(_MonsterRefine):
    # placement dependent methods

    def partial_blending(self) -> None:
        """
        multiple possible scaffolds for placement and best is chosen
        """
        self.mol_options = self.partially_blend_hits()  # merger of hits
        unrefined_scaffold, mode_index = self.pick_best()
        used = unrefined_scaffold.GetProp('_Name').split('-')
        self.unmatched = [h.GetProp('_Name') for h in self.hits if h.GetProp('_Name') not in used]
        scaffold = self.posthoc_refine(unrefined_scaffold)
        chimera = self.make_chimera(scaffold, mode_index)

        # atom_map is filled via `self.get_mcs_mapping(target_mol, template_mol)`:
        self.positioned_mol = self.place_from_map(target_mol=self.positioned_mol,
                                                  template_mol=chimera,
                                                  atom_map=None,
                                                  random_seed=self.random_seed)
        self.keep_copy(scaffold, 'scaffold')
        self.keep_copy(chimera, 'chimera')

    def partially_blend_hits(self, hits: Optional[List[Chem.Mol]] = None) -> List[Chem.Mol]:
        """
        This is the partial merge algorithm, wherein the hits are attempted to be combined.
        If the combination is bad. It will not be combined.
        Returning a list of possible options.
        These will have the atoms changed too.

        :param hits:
        :param distance:
        :return:
        """

        if hits is None:
            hits = sorted(self.hits, key=lambda h: h.GetNumAtoms(), reverse=True)
        # this predates fix_hits. Check if it is still needed.
        for hi, hit in enumerate(hits):
            # fallback naming.
            if not hit.HasProp('_Name') or hit.GetProp('_Name').strip() == '':
                hit.SetProp('_Name', f'hit{hi}')

        ## a dodgy hit is a hit with inconsistent mapping between three.
        def get_dodgies(skippers):
            dodgy = []
            for hit0, hit1, hit2 in itertools.combinations(hits, 3):
                hn0 = hit0.GetProp('_Name')
                hn1 = hit1.GetProp('_Name')
                hn2 = hit2.GetProp('_Name')
                if any([hit in skippers for hit in (hn0, hn1, hn2)]):
                    continue
                for a, b in inter_mapping[(hn0, hn1)].items():
                    if a in inter_mapping[(hn0, hn2)] and b in inter_mapping[(hn1, hn2)]:
                        if inter_mapping[(hn0, hn2)][a] != inter_mapping[(hn1, hn2)][b]:
                            # TODO: THIS IS A BAD OPTION:
                            # if all([m.GetAtomWithIdx(i).IsInRing() for m, i in ((hit0, a),
                            #                                                     (hit1, b),
                            #                                                     (hit2, inter_mapping[(hn0, hn2)][a]),
                            #                                                     (hit2, inter_mapping[(hn1, hn2)][b]))]):
                            #     pass
                            # else:
                            dodgy.extend((hn0, hn1, hn2))
            d = Counter(dodgy).most_common()
            if dodgy:
                return get_dodgies(skippers=skippers + [d[0][0]])
            else:
                return skippers

        inter_mapping = {}
        for h1, h2 in itertools.combinations(hits, 2):
            inter_mapping[(h1.GetProp('_Name'), h2.GetProp('_Name'))] = self.get_positional_mapping(h1, h2)
        dodgy_names = get_dodgies([])
        warn(f'These combiend badly: {dodgy_names}')
        dodgies = [hit for hit in hits if hit.GetProp('_Name') in dodgy_names]
        mergituri = [hit for hit in hits if hit.GetProp('_Name') not in dodgy_names]
        merged = self.simply_merge_hits(mergituri)
        dodgies += [hit for hit in hits if hit.GetProp('_Name') in self.unmatched]
        self.unmatched = []
        combined_dodgies = []
        for h1, h2 in itertools.combinations(dodgies, 2):
            h_alt = Chem.Mol(h1)
            try:
                combined_dodgies.append(self.merge_pair(h_alt, h2))
            except FragmensteinError:
                pass
        combinations = [merged] + dodgies + combined_dodgies
        # propagate alternatives
        while self.propagate_alternatives(combinations) != 0:
            pass
        return combinations

    def pick_best(self) -> Tuple[Chem.Mol, int]:
        """
        Method for partial merging for placement

        :return: unrefined_scaffold, mode_index
        """
        if len(self.mol_options) == 1:
            return self.mol_options[0], 0
        elif len(self.mol_options) == 0:
            raise ValueError('No scaffolds made?!')
        else:
            mapx = {}  #: dictionary of key mol name and value tuple of maps and mode

            def template_sorter(t: List[Chem.Mol]) -> float:
                # key for sorting. requires outer scope ``maps``.
                self.journal.debug(f'pick_best (partial merging): Sorting {len(t)} templates: {t}')
                n_atoms = len(mapx[t[0].GetProp('_Name')])
                mode = mapx[t[1].GetProp('_Name')]
                mode_i = self.matching_modes.index(mode)
                return - n_atoms - mode_i / 10

            ## get data
            # presort as this is expensive.
            for template in self.mol_options:
                # _get_atom_maps returns a list of alternative mappings which are lists of template to initail mol
                # todo flip around
                atom_maps: List[IndexMap] = self._get_atom_maps(followup=self.initial_mol, hit=template,
                                                                atomCompare=rdFMCS.AtomCompare.CompareElements,
                                                                bondCompare=rdFMCS.BondCompare.CompareOrder,
                                                                ringMatchesRingOnly=True,
                                                                ringCompare=rdFMCS.RingCompare.PermissiveRingFusion,
                                                                matchChiralTag=False)
                atom_maps: List[IndexMap] = flip_mapping(atom_maps)
                mapx[template.GetProp('_Name')] = (atom_maps, self.matching_modes[-1])
            # search properly only top 3.
            self.mol_options = sorted(self.mol_options, key=template_sorter)
            for template in self.mol_options[:3]:
                atom_map, mode = self.get_mcs_mapping(template, self.initial_mol)
                # get_mcs_mapping returns a dict going from template index to initial.
                mapx[template.GetProp('_Name')] = (atom_map, mode)
                self.journal.debug(f"With {template.GetProp('_Name')}, " + \
                                   "{len(atom_map)} atoms map using mode {self.matching_modes.index(mode)}")
            ## pick best template
            self.mol_options = sorted(self.mol_options, key=template_sorter)
            ## Check if missing atoms can be explained by a different one with no overlap
            best = self.mol_options[0]
            ## Fuse overlaps
            # best_map = maps[best.GetProp('_Name')][0]
            # full = set(range(self.initial_mol.GetNumAtoms()))
            # present = set(best_map.values())
            # missing = full - present
            # for other in self.mol_options:
            #     other_map = maps[other.GetProp('_Name')][0]
            #     found = set(other_map.values())
            #     if len(found) > 6 and len(present & found) == 0: # more than just a ring and no overlap
            #         fusion = self._fuse(best, other, best_map, other_map)
            return best, self.matching_modes.index(mapx[best.GetProp('_Name')][1])

    # def _fuse(self, mol_A: Chem.Mol, mol_B: Chem.Mol, map_A: Dict[int, int], map_B: Dict[int, int]) -> Chem.Mol:
    #     """
    #     Merge two compounds... but that are unlinked, using the followup as a guide.
    #     Conceptually different but overlapping is join_neighboring_mols
    #
    #     :param mol_A:
    #     :param mol_B:
    #     :param map_A:
    #     :param map_B:
    #     :return:
    #     """
    #     # No longer needed.
    #     fusion = Chem.RwMol(Chem.CombineMols(mol_A, mol_B))
    #     t = mol_A.GetNumAtoms()
    #     new_map_B = {k+t: v for k, v in map_B.items()}
    #     full = set(range(self.initial_mol.GetNumAtoms()))
    #     present_A = set(map_A.values())
    #     present_B = set(map_B.values())
    #
    #     def find_route(n):
    #         if n in present_A:
    #             return None
    #         elif n in present_B:
    #             return n
    #         else:
    #             path_raw = {m: find_route(m) for m in self.initial_mol.GetAtomWithIdx(n).GetNeighbors()}
    #             path = {i: path_raw[i] for i in path_raw if path_raw[i] is not None}
    #             if len(path) == 0:
    #                 return None
    #             else:
    #                 return {n: path}



from .base import FragmensteinParserBase, _add_common, set_verbose
import argparse
from fragmenstein import Monster
from rdkit import Chem

class FragmensteinParserMonster:

    def _define_monster(self, parser: argparse.ArgumentParser):
        """fragmenstein monster combine -i hit1.mol hit2.mol
           fragmenstein monster place -i hit1.mol hit2.mol -s 'CCO'"""
        subparsers = parser.add_subparsers(title='operation', help='combine or place')
        combine_parser = subparsers.add_parser('combine', help='combine')
        _add_common(combine_parser, hits=True, output=False, name=False)
        combine_parser.set_defaults(func=self.monster_combine)
        place_parser = subparsers.add_parser('place', help='place')
        place_parser.add_argument('-s', '--smiles', required=True)
        _add_common(place_parser, hits=True, output=False, name=True)
        place_parser.set_defaults(func=self.monster_place)

    def monster_combine(self, args: argparse.Namespace) -> str:
        set_verbose(args.verbose)
        monsta = Monster(hits=list(map(Chem.MolFromMolFile, args.hits)))
        monsta.combine()
        return Chem.MolToMolBlock(monsta.positioned_mol)

    def monster_place(self, args: argparse.Namespace) -> str:
        set_verbose(args.verbose)
        monsta = Monster(hits=list(map(Chem.MolFromMolFile, args.hits)))
        monsta.place_smiles(args.smiles, long_name=args.name)
        return Chem.MolToMolBlock(monsta.positioned_mol)



# these are not tests. But cases as a human I would not know how to solve.

def ring_mess():
    """
    This is a case of two rings that are close and result in a weird molecule due to overzealous absorption.
    """
    hits = [Mac1.get_mol('diamond-x0711_A'), Mac1.get_mol('diamond-x0689_A')]
    Victor.monster_throw_on_discard = True
    Victor.quick_reanimation = True
    victor = Victor(hits=hits,
                    pdb_block=Mac1.get_template(),
                    )
    victor.combine()
    mol = victor.monster.modifications['expanded']

    from molecular_rectifier import Rectifier

    recto = Rectifier(mol)
    recto.fix()
    # recto.modifications[3] is good as gold
    # recto.modifications[4] is a mess


__version__ = "0.12.8"

from typing import Dict
def get_versions() -> Dict[str, str]:
    """
    Return a dict of versions of os, python, fragmenstein etc.
    """
    import pkg_resources, sys, platform

    get_version = lambda name: pkg_resources.get_distribution(name).version

    return dict(python=sys.version, os_type=platform.system(), arc=platform.machine(),
                fragmenstein=get_version("fragmenstein"),
                pyrosetta=get_version("pyrosetta"),
                rdkit=get_version("rdkit"),
                molecular_rectifier=get_version('molecular-rectifier'),
                rdkit_to_params=get_version('rdkit-to-params')
                )



"""
Victor alternatives not meant for production use, but for testing and debugging.

* ``Mictor``: MCS hack for negative benchmarking: Victor with MCSMerger, i.e. no positional info used.
* ``Wictor``: Victor without pyrosetta, i.e. no energy minimization.
* ``AccountableBRICS``: Not a Victor, but generates a BRICS decomposition of the hits and returns a dataframe
    with the built molecules usable in ``Laboratory.place``.
* ``FreeVictor``: Victor with no constraints.
* ``SingleVictor``: Victor with only one hit used for constraints.

To run in ``Laboratory`` class set the ``.Victor`` attribute to the desired ``Victor`` alternative.

... code-block:: python
    from fragmenstein import Laboratory
    from fragmenstein.faux_victors import Mictor  # MCS hack for negative benchmarking
    lab = Laboratory(pdbblock=pdbblock, covalent_resi=None)
    lab.Victor = Mictor
    combinations: pd.DataFrame = lab.combine(hits, n_cores=28)
    combinations.to_pickle(f'👾👾👾.p'))
"""

from .mcs_victor import Mictor, MCSMerger
from .no_pyrosetta import Wictor
from .brics import AccountableBRICS
from .others import FreeVictor, SingleVictor
from .quick import Quicktor
